;==============================================================
;Note: Case insensitive
	LIST P=PIC16F84
  	ERRORLEVEL -302,-305
;
	include <p16f84.inc>
;
;       Title:  caliper6.asm
;       Authors: Lee Studley & Wilfried Fedtke 
;       Edit_date: 3/27/2003 as of
;
;*****Configuration for 16F84 watchdog on
	__CONFIG   _CP_OFF & _WDT_ON & _PWRTE_ON & _HS_OSC 
;
; History:
; cleanup to support schematic serial only
; Supports Serial 57600 baud 8N1
;
;==============================================================
;====DEFINES===================================================
;==============================================================
;**************************************************************
;*	----------------------------------	
;*	PORTA: 	0 - 	
;*		1 -	
;*		2 -	serial output :	
;*		3 -	Caliper buffered Clk in
;*		4 -	Caliper buffered Data in
;*	PORTB:	0 	serial input/INT (TTL)	not used
;
#define Serial_bit     PORTA,0x2      ; serial output (TTL)if 16C558 jmpered to portC:6
#define Serialin_bit   PORTB,0x0      ; serial input/INT (TTL)
;
#define gp_ramstart    0x0C        ;F84 start of user register/variables 68 lcations
#define gp_ramend      0x4F        ;F84 end of user register/variables
;==============================================================
;====PROGRAM DATA==============================================
;==============================================================
	cblock gp_ramstart      ; Reserve/define register variables
	w_sav                   ; Remember to reserve also in other banks too 
	status_sav              ;  
	fsr_sav                 ;  
	pclath_sav              ;
	gP_Flags    ; Global general purpose flags
;===========================================================================
;===STACK and CONTEXT ======================================================
;===========================================================================
; A software Stack is implemented by calling the routines 'Push' or 'Pop' 
; with the address of the variable to be operated on in w. The stack pointer 
; at reset is initialized to the last ram location,(gp_ramend) and grows upward. 
; Uses: StackTmp,StackFSR,pStack,w.
;==========================================================================
	StackFSR                ; Stack routine Variable to save orig FSR 1st
	StackTmp                ; Stack Scratch Var
	pStack                  ; Variable pointing to the current Var on stack 
	Int_Shadows:3     ; Shadows for int manager variables
;==========================================================================
; Push/PopVars: gives a ~fast(34uS@4mhz) context save within the interrupt 
; handler for vars temp0..15. Uses only w (already backed up by the standard
; interrupt mechanism into w_sav). Frees up all temps to use in routines
;==========================================================================
; Volatile General scratch/temp variables for routines
;==========================================================================
	temp0                   ; total of xx registers accessed by "temp0+x"
	temp1                   ;
	temp2                   ;
	temp3                   ;
	temp4                   ;
	temp5                   ;
	temp6                   ;
	temp7                   ;
	temp8                   ; 
	temp9                   ;
	temp10                  ;
	temp11                  ;
	temp12			;
	temp13:10               ; 
	endc                    ; End of constants/labels
;===============================;
; gP_Flags general purpose flag Equates:
SignFlag EQU 0x7    ; caliper data sign flag
;  EQU 0x6    ; n/a
;  EQU 0x5    ; n/a
;  EQU 0x4    ; n/a
;  EQU 0x3    ; n/a
;  EQU 0x2    ; n/a
;  EQU 0x1    ; n/a
;  EQU 0x0    ; n/a
;==========================================================================
;


;==============================================================
;====PROGRAM CODE==============================================
;==============================================================
	org     0x00
Reset   clrf    STATUS          ;
	clrf    PCLATH          ;
	clrf    INTCON          ;
	goto    cold_init       ;

;==============================================================
;========== Interrupt handler: saves w,STATUS,FSR,PCLATH=======
;==============================================================
	org 0x04
	retfie
;==============================================================
;==============================================================
;==============================================================

cold_init        
	clrf    FSR
	movlw   b'11111111'     ; 0xFF (bit 6 is Tx on Picdem2)
	movwf   PORTA    		; init to Tx=resting=1
	movlw   gp_ramend       ; get stack bottom
	movwf   pStack          ; init stack pointer

warm_init
;============== periph and general setup======================
	clrf    TMR0 
	bsf     STATUS,RP0      ; set to page1 register access 
	movlw   b'00001111'     ; set prescaler to wdt & 1/128 15msx128=~2sec wdt
	movwf   OPTION_REG
	movlw   b'11111011'     ; 0xFB only output is Tx so far
	movwf   TRISA             
	movlw	B'00000000'		; PortB all outputs
	movwf   TRISB             
	bcf     STATUS,RP0      ; set back to page0  
	movlw   b'00000000'     ; turn off all interrupts for now
	movwf   INTCON      ; Set INTE=1 RB0/int enabled
;============== end periph and general setup===+===============


;==============================================================
;==============================================================
;==============================================================
;=Main Loop====================================================
;==============================================================
Main    
	clrwdt      ; if main is not executed every ~3sec reset
	movlw 0x00
	movwf temp1
	movlw 0x0F
	movwf temp2

;==============================================================
;===Get Caliper data===========================================
loops_o btfsc PORTA,0x3 ; CLK loop till lo for long time
			clrf temp1  ; if hi, loop and reset temp1 count
loops_i decfsz temp1    
	goto loops_o
	decfsz temp2
	goto loops_o
;==============================================================
; Take advantage of pause between caliper 24bit bursts of data by using 2
; 24 bit timing measurements. This way only 3 registers need rrf's in each
; loop (as opposed to 48bit: 6 rrf's) which would slow the timing down too 
; much for a 90khz period measurement.
; clock stays high during pause _____xxxxxx------xxxxxxx____ clk
; Example: inch 1.000 == FFAFFFFF9D41 ==> FFAFFF FF9D41 => 1.000 (2'scmp encoded)
; 1st 24bits is relative(same as display)   and 2nd 24bits is absolute
; output: data in temp1..temp6
;==================

cloop_proceed
	movlw d'24'
	movwf temp12
	clrf temp1     ; msb of 1st 24bit section
	clrf temp2
	clrf temp3
	clrf temp4     ; msb of 2nd abs 24bit section
	clrf temp5
	clrf temp6
;
cloop   ; 'capture' loop
	clrf STATUS 
	btfss PORTA,0x3 ; CLK loop till hi
	goto $-1
	btfsc PORTA,0x3 ; CLK loop till lo (falling edge detect)
	goto $-1
	btfsc PORTA,0x4 ; data SDA test the data line
	bsf STATUS,C        
	rrf temp4   ; 6 byte wide shift register    
	rrf temp5   ; shifting(rotating)right thru carry
	rrf temp6
	decfsz temp12   ; when 24 bits are done, skip 
	goto cloop  ; <48 so loop again
;
; caliper pause alittle before next burst of 24bits sent
; clock stays high during pause _____xxxxxx------xxxxxxx____ clk
;                                    24bit       24bit
; get 2nd set of 24bits
;
	movlw d'24'
	movwf temp12

cloop2  ; 2nd 'capture' loop
	clrf STATUS 
	btfss PORTA,0x3 ; CLK loop till hi
	goto $-1
	btfsc PORTA,0x3 ; CLK loop till lo (falling edge detect)
	goto $-1
	btfsc PORTA,0x4 ; data SDA test the data line
	bsf STATUS,C        
	rrf temp1   ; 6 byte wide shift register    
	rrf temp2   ; shifting(rotating)right thru carry
	rrf temp3
	decfsz temp12   ; when 24 bits are done, skip 
	goto cloop2  ; <48 so loop again
;=======Endof-Get Caliper data=================================
;==============================================================

;=====================================================
;1) test if 2's comp
;    yes? set sign flag to "+"
;    no? is neg# so 2's comp it
;2) mult by 1000dec 
;3) divide by 5000h
;4) call 32binary-10digAscii
;5) rearrange and add +/- char and decimal point
;Equations:
; 	mm*100 = ( sample*1000) / 8063 (806.2992126 * 10)  
; 	mm*100 = (sample*0x03E8) / 0x1F7F (806.2992126 * 10)  
;	inches*1000 = (sample*1000)/20480

isPosorNeg
	bsf   gP_Flags,SignFlag    ; set flag default to indicate a caliper negative distance  "-"      
	btfss temp1,0x7   ; NOTE: calip data is 2's comp if POSITIVE , normal if negative (Backward from intuitive)
	goto  nowisabs ; data is positive so'caliper distance is negative'      
	call  cmp2s    ; MSB is set so 'caliper distance is positive so do a 2's complement on 24bits   
	bcf   gP_Flags,SignFlag    ; set flag to indicate a positive distance   "+"      
;
nowisabs 
	;==setup for MULTIPLY_24x24 multiply routine================
	clrf temp7     ; NOTE: See equates in call routine
	movlw 0x03     ;001000dec = 0x0003E8    24bit
	movwf temp8
	movlw 0xE8
	movwf temp9       
	call  MULTIPLY_24x24 ; product will be in temp4,5,6,7,8,9 6bytes total when done
;
;==setup for divide routine:==========================================
;Divisor is in temp0..temp2 (3bytes 24bits)
	clrf  temp0; is =0 for both options
; TODO put inch/mm flag test here
inch_load
	movlw 0x50 ; load divisor with 0x5000(20480 cnts/inch)
	movwf temp1 
	clrf temp2 
	goto doDivide;

mm_load
	movlw 0x1F ; load divisor with 0x1F7F(8063 cnts/mm)
	movwf temp1
	movlw 0x7F  
	movwf temp2 
;
doDivide
	call  DIVIDE_48by24  ; result in temp4,5,6,7,8,9
;
;==Setup for binary to ascii routine:===================================
; temp0,1,2,3 32-bit binary number (unsigned)  msb-lsb order bin..bin+3
; result in temp4,5,6,7,8,9,10,11,12,13     10 BC digits or 10 ascii chars
	movf temp6,w
	movwf temp0	
	movf temp7,w
	movwf temp1	
	movf temp8,w
	movwf temp2	
	movf temp9,w
	movwf temp3	
	call b2bcd     ; convert result to 32-bit binary to 10 bcd
	call bcd2a     ; convert result 10 bcd to 10 ascii
;
;==Rearrange and add "+/-" and  decimal point Format Output=============
; temp0+4..temp0+13 have the ascii output
	;  tens ascii digit
	movf  temp0+10,w   ; if was not = 0 retreive it and stuff it
	movwf temp0+18    ; stuff the tens MM digit place
	;  hundreds ascii digit
	movf  temp0+9,w   ; if was not = 0 retreive it and stuff it
	movwf temp0+17    ; stuff the tens MM digit place
	;  thousands ascii digit
	movf  temp0+8,w   ; if was not = 0 retreive it and stuff it
	movwf temp0+16    ; stuff the tens MM digit place
	;  10 thousands ascii digit
	movf  temp0+7,w   	   ; if was not = 0 retreive it and stuff it
	movwf temp0+15 		   ; stuff the tens MM digit place
	movf  temp0+6,w   	   ; if was not = 0 retreive it and stuff it
	movwf temp0+14 		   ; stuff the tens MM digit place
	movf  temp0+13,w ; thousandths ascii digit (LSDigit)
	movwf temp0+22
	movf  temp0+12,w ; hundredths ascii digit
	movwf temp0+21

; TODO put inch/mm flag test here
;	goto	FormatOutputMM
	goto	FormatOutputInches
;
FormatOutputInches
	movf  temp0+11,w ; tenths ascii digit
	movwf temp0+20
	movlw '.'   ; ascii decimal point 
	movwf temp0+19
	goto killleadzeros
;
FormatOutputMM
	movlw '.'   ; ascii decimal point 
	movwf temp0+20
	movf  temp0+11,w ; ones MM ascii digit
	movwf temp0+19
;
killleadzeros
	movlw temp0+14-1 	; set to 1 before area we are interested in 
	movwf FSR
;
tstforZ	movlw '0'
	incf FSR
	xorwf INDF,w
	btfss STATUS,Z
	goto stuffsign
	movlw ' '
	movwf INDF
	goto tstforZ	
stuffsign
	movlw '.'
	xorwf INDF,w
	btfss STATUS,Z
	goto noleadzero
	decf FSR
	movlw '0'   
	movwf INDF	; put a single lead zero back if needed
noleadzero
	decf FSR
	; test for ' ' or - sign
	movlw ' '   ; default is positive (no "-" sign)
	btfsc gP_Flags,SignFlag ; add the "-" sign if negative flag set
	movlw '-'   
	movwf INDF
;==============================================================
;Send position data out the serial pin in ascii-hex format 
	movlw FSR   ; Save off the current FSR  
	call Push   
	movlw temp0+14-1    ; get address preceeding the buffer
	movwf FSR           ; put it into indirect pointer FSR
	movlw 9         	; byte count       
	movwf temp0+30      ; use temp0+20 as counter
;
SendDataSerial
	incf FSR        ; pre increment FSR to point to temp1   
	movf INDF,w    ; get the register contents pointed to by FSR via INDF
	movwf CharStor
	call Serial     ; send char out CharStor byte via serial pin
	decfsz temp0+30
	goto SendDataSerial
;
; finished sending 6 ascii-hex bytes serially
	movlw 0x0A    ; *** \n newline
	movwf CharStor
	call  Serial  ; echo char via rs232
	movlw 0x0D    ; *** return
	movwf CharStor
	call  Serial  ; echo char via rs232
; sent a newline & return
	movlw FSR   ; restore the 'Pushed' FSR  
	call Pop    
	goto Main           
;	
;============================================================
;====SUBROUTINES=============================================
;============================================================

;============================================================
; Upon entry, the desired pin must already be set up as an output and 
; to the Idle state. 
; uses delayCnt,SerialCnt, CharStor=xmit byte
CharStor        equ     temp9
SerialCnt       equ     temp10
delayCnt        equ     temp11
;=========
clockratex      equ     d'10000000'    ;define Fosc rate here
baudratex       equ     d'57600'        ;define baud rate here
Iclkx           equ     clockratex/4   ;Instruction rate
baud_val        equ     (((Iclkx/baudratex)-d'12')/d'13')  ;8bit value
; example hard-code values:
;#define baud_val       d'79'           ; 2400 @ 10mhz
;#define baud_val       d'19'           ; 9600 @ 10mhz
;#define baud_val       d'31'           ; 2400 @ 4mhz
;#define baud_val       d'31'           ; 9600 @ 16mhz
;============================================================
Serial  movlw   SerialCnt   ; get addr of var to save/restore
	call    Push        ; Push/Pop to software stack
	movlw   delayCnt    ; get addr of var to save/restore
	call    Push        ; Push/Pop to software stack
	bcf     STATUS,C    ; Start bit
	call    bitout
	movlw   0x08
	movwf   SerialCnt   ; number of bits to shift
bits    rrf     CharStor    ;   
	call    bitout      ;
	decfsz  SerialCnt       ;
	goto    bits        ;
	rrf     CharStor    
	bsf     STATUS,C    ; 1 Stop bit
	call    bitout

	movlw   delayCnt    ; get addr of var to save/restore
	call    Pop         ; Push/Pop to software stack
	movlw   SerialCnt   ; get addr of var to save/restore
	call    Pop         ; Push/Pop to software stack
	return  ;********Exit point
bitout  btfss   STATUS,C    ; NOTE: some RS232 cards can accept 0-5v RS232  
	bcf     Serial_bit  ; --Use if Max232 chip  
	btfsc   STATUS,C    
	bsf     Serial_bit  ; --Use if Max232 chip
	movlw   baud_val
dly 
	goto    $+1 ;**optimized at 57600@10mhz max ~17.4us/bit
	nop ; *****optimized at 57600@10mhz max( slower rates still work)
	movwf   delayCnt
dloop
	goto    $+1
	goto    $+1
	goto    $+1
	goto    $+1
	goto    $+1
	decfsz  delayCnt
	goto    dloop
	return
;============================================================
;binasci: Enter with byte to be converted in w, returns converted in 
;w(msbs ascii) and AsciVar(lsb ascii)
;routine to convert hex to ascii 
;uses AsciVar(corrupts)
;============================================================
AsciVar equ     temp8
;============================================================
binasci 
	movwf   AsciVar           ; put copy into AsciVar
	andlw   0x0F            ; mask off upper nibble
	addlw   0xF6            ; test to see if 0-9 is in W
	btfsc   STATUS,C         
	addlw   0x07            ; if >9, then add to get to A-F range
	addlw   0x3A            ; add 0Ah and 30h ascii offeset back in
	swapf   AsciVar           ; (msbs)swap and get nibbles
	xorwf   AsciVar           ; swap AsciVar and W
	xorwf   AsciVar,w         ;
	xorwf   AsciVar           ; now AsciVar=lsbs(ascii), w=msbs(hex)
	andlw   0x0F            ; mask off upper nibble
	addlw   0xF6            ; test to see if 0-9 is in W
	btfsc   STATUS,C         
	addlw   0x07            ; if >9, then add to get to A-F range
	addlw   0x3A            ; add 0Ah and 30h ascii offeset back in
	return
;============================================================
; Pause: Upon entry, variables CntTmpA and CntTmpB hold the 
; MSB and LSB of the number of milliseconds 1..65535 delay.
; uses CntTmpA,CntTmpB,CntTmpC(corrupts all) @ 4mhz
;============================================================
CntTmpA equ     temp0+14        ; Msbs       
CntTmpB equ     temp0+15        ; Lsbs
CntTmpC equ     temp0+16        ; scratch counter
;============================================================

pause   movlw   0x00
	movwf   CntTmpA
	movlw   d'20'
	movwf   CntTmpB
NmSdelay 
	comf    CntTmpA
	comf    CntTmpB
	incf    CntTmpB
	btfsc   STATUS,Z
	incf    CntTmpA
load    movlw   0xF8
	movwf   CntTmpC
eloop   nop
	decfsz  CntTmpC
	goto    eloop
	goto    $+1
	incf    CntTmpB
	btfsc   STATUS,Z
	incfsz  CntTmpA
	goto    load
	return   
;============================================================
RetNewline      ;
	; Also other routines call it to do a "\n"
	movlw   0x0A    ; *** \n newline
	movwf   CharStor
	call    Serial  ; echo char via rs232
	movlw   0x0D    ; *** return
	movwf   CharStor
	call    Serial  ; echo char via rs232
	retlw   1       ; return
;==========================================================================
; Push/Pop: Software stack(~TBDuS/bit@4mhz) 
; Entry: with addr of Var to push in w. Original FSR is unchanged
; Exit: pStack will point to the last Var on the stack
; Uses: StackTmp,StackFSR,pStack,w
;==========================================================================
; Use:  movlw temp6     ; Save temp6 on top of stack
;       call  Push
;==========================================================================
Push    movwf   StackTmp        ; Save the addr in StackTmp
	movf    FSR,w           ; Get current FSR
	movwf   StackFSR        ; Save it in StackFSR
	movf    StackTmp,w      ; Get addr again
	movwf   FSR             ; Init FSR to point there
	movf    INDF,w          ; Get the Var to push
	movwf   StackTmp        ; Save the Var in Tmp
	movf    pStack,w        ; Get pointer
	movwf   FSR             ; Init FSR w/ the stack pointer    
	movf    StackTmp,w      ; Get the Var again
	movwf   INDF            ; put it on the stack
	decf    pStack          ; Put Stack pointer to the Var just stacked
	movf    StackFSR,w      ; Get original  FSR
	movwf   FSR             ; Restore it
	retlw   1        
;==========================================================================
; Entry with ADDR pf Var to pop in w. Original FSR is unchanged
; Exit: pStack will point to the last Var left on the stack
; Uses: StackTmp,StackFSR,pStack,w
;==========================================================================
; Use:  movlw temp6     ; restore temp6 from top of stack
;       call  Pop
;==========================================================================
Pop     movwf   StackTmp        ; Save the addr in StackTmp
	movf    FSR,w           ; Get current FSR
	movwf   StackFSR        ; Save it in StackFSR
	incf    pStack          ; Put Stack pointer to last pushed Var
	movf    pStack,w        ; Get stack pointer
	movwf   FSR             ; Init FSR w/ the stack pointer    
	movf    INDF,w          ; Get Var from the stack
	xorwf   StackTmp        ; Swap Var into Tmp
	xorwf   StackTmp,w      ; Get ADDR to pop in W
	xorwf   StackTmp        ; 
	movwf   FSR             ; Init FSR to point to Var to pop
	movf    StackTmp,w      ; get Var from Tmp
	movwf   INDF            ; pop the Var back to ADDR
	movf    StackFSR,w      ; Get original  FSR
	movwf   FSR             ; Restore it
	retlw   1        

;=======================================================================
cmp2s   ; do a 2's complement on 24bits     
	comf    temp0+3       ;Complement all bytes
	comf    temp0+2
	comf    temp0+1
	incf    temp0+3       ;Inc. low byte always
	skpnz                 ;Skip if no carry to higher bytes
	incf    temp0+2       ;Carry to next byte
	skpnz                 ;Skip if no carry to higher bytes
	incf    temp0+1          ;Carry to next byte
	return
;=======================================================================
;******************************************************************
;title   "24x24bit multiply"
;MULTIPLY_24x24
;24 x 24 Multiplication
;Input:
; Multiplier - 3 bytes (shared with Product)  (msb,lsb)
; Multiplicand - 3 bytes (not modified)       (msb,lsb)
;Temporary:
; Bitcount
;Output:
; Product - 6 bytes  (msb,lsb)
;******************************************************************
; cblock 0x0c
;         Product:6
;         Multipland:3
;         BitCount:1
;        endc

Multipland EQU temp1      ;temp1,2,3
Product    EQU temp4      ;temp4,5,6
Multiplier EQU Product+3  ;temp7,8,9 : 3 bytes shared with Product's
			;less significant bytes (+3..5)
BitCount   EQU temp0 
;=====================================================

MULTIPLY_24x24
	CLRF    Product         ; clear destination
	CLRF    Product+1
	CLRF    Product+2
			
	MOVLW   D'24'
	MOVWF   BitCount        ; number of bits
	RRF     Product+3,F     ; shift out to carry
	RRF     Product+4,F     ; next multiplier bit
	RRF     Product+5,F

ADD_LOOP_24x24

	BTFSS   STATUS,C        ; if carry is set we must add multipland
									; to the product
	GOTO  SKIP_LOOP_24x24 ; nope, skip this bit
		  
	MOVF    Multipland+2,W  ; get LSB of multiplicand
	ADDWF   Product+2,F     ; add it to the lsb of the product

	MOVF    Multipland+1,W  ; middle byte
	BTFSC   STATUS,C        ; check carry for overflow
	INCFSZ  Multipland+1,W  ; if carry set we add one to the source 
	ADDWF   Product+1,F     ; and add it  (if not zero, in
								; that case mulitpland = 0xff->0x00 )

	MOVF    Multipland,W    ; MSB byte
	BTFSC   STATUS,C        ; check carry
	INCFSZ  Multipland,W
	ADDWF   Product,F       ; handle overflow

SKIP_LOOP_24x24
	; note carry contains most significant bit of
	; addition here

	; shift in carry and shift out
	; next multiplier bit, starting from less
	; significant bit

	RRF     Product,F
	RRF     Product+1,F
	RRF     Product+2,F
	RRF     Product+3,F
	RRF     Product+4,F
	RRF     Product+5,F

	DECFSZ  BitCount,F
	GOTO    ADD_LOOP_24x24
	RETURN

;==========================================================================
;==========================================================================
;***********************************************************************
; DIVIDE_48by24 - Routine to divide a 48 bit number with a 24 bit number
; result upto 48 bits !
;
; Formula:      Dividend = Dividend/Divisor
;       
; Format:       Little endian, Ram = msb, Ram+x = lsb ( where x=bytesize-1)
; Ram used:     Dividend 6 bytes ( 48 bits )
;               Divisor 3 bytes ( 24 bits )
;               Temp 4 bytes 
;               BitCount 1 byte for loop counting ( and temp. saving of carry )
; 
; Divisor is preserved
; Dividend holds result
; Returns with zero in W if failed ( division with zero )
; Else returns with one in w
; 
; Based on pseudo-code from Nikolai Golovchenko [golovchenko@mail.ru]
;*********************************************************************** 
Dividend	EQU temp4	; 6bytes (48bits)
Divisor		EQU temp0	; 3 bytes(24bits
BitCnt		EQU temp12 ;
Tmp		EQU  temp0+13	;	

DIVIDE_48by24
  ; Test for zero division
	MOVF    Divisor,W
  	IORWF   Divisor+1,W
	IORWF   Divisor+2,W
	BTFSC   STATUS,Z
	RETLW   0x00    ; divisor = zero, not possible to calculate return with zero in w

  ; prepare used variables
	CLRF    Tmp
	CLRF    Tmp+1
	CLRF    Tmp+2

	clrf    Tmp+3

	MOVLW   D'48'           ; initialize bit counter
	MOVWF   BitCnt

DIVIDE_LOOP_48by24
	RLF     Dividend+5,F
	RLF     Dividend+4,F
	RLF     Dividend+3,F
	RLF     Dividend+2,F
	RLF     Dividend+1,F
	RLF     Dividend,F
  ; shift in highest bit from dividend through carry in temp
	RLF     Tmp+2,F
	RLF     Tmp+1,F
	RLF     Tmp,F

	rlf     Tmp+3, f

	MOVF    Divisor+2,W     ; get LSB of divisor
	btfsc   Tmp+3, 7
	goto    Div48by24_add

  ; subtract 24 bit divisor from 24 bit temp
	SUBWF   Tmp+2,F        ; subtract

	MOVF    Divisor+1,W     ; get middle byte
	SKPC                    ;  if overflow ( from prev. subtraction )
	INCFSZ  Divisor+1,W     ; incresase source
	SUBWF   Tmp+1,F        ; and subtract from dest.

	MOVF    Divisor,W       ; get top byte
	SKPC                    ;  if overflow ( from prev. subtraction )
	INCFSZ  Divisor,W       ; increase source
	SUBWF   Tmp,F          ; and subtract from dest.

	movlw 1
	skpc
	subwf   Tmp+3, f
	GOTO    DIVIDE_SKIP_48by24 ; carry was set, subtraction ok, continue with next bit

Div48by24_add
  ; result of subtraction was negative restore temp
	ADDWF   Tmp+2,F        ; add it to the lsb of temp

	MOVF    Divisor+1,W     ; middle byte
	BTFSC   STATUS,C        ; check carry for overflow from previous addition
	INCFSZ  Divisor+1,W     ; if carry set we add 1 to the source
	ADDWF   Tmp+1,F        ; and add it if not zero in that case Product+Multipland=Product

	MOVF    Divisor,W       ; MSB byte
	BTFSC   STATUS,C        ; check carry for overflow from previous addition
	INCFSZ  Divisor,W
	ADDWF   Tmp,F          ; handle overflow

	movlw 1
	skpnc
	addwf   Tmp+3, f

DIVIDE_SKIP_48by24
	DECFSZ  BitCnt,F      ; decrement loop counter
	GOTO    DIVIDE_LOOP_48by24      ; another run
  ; finally shift in the last carry
	RLF     Dividend+5,F
	RLF     Dividend+4,F
	RLF     Dividend+3,F
	RLF     Dividend+2,F
	RLF     Dividend+1,F
	RLF     Dividend,F
	RETLW   0x01    ; done

;******************************************************************
;
; 32-bit unsigned binary to BCD and BCD to ASCII.
;
;******************************************************************
;        title   "32-bit binary to ascii"
; CBLOCK   temp1;
; bin:4    ; 32-bit binary number (unsigned)  msb-lsb order bin..bin+3
; bcd:10      ; 10 BC digits or 10 ascii chars
; pti,pto     ; pointers
; ii
; temp
; cnt
; ENDC

bin  EQU temp0      ; temp0,1,2,3 32-bit binary number (unsigned)  msb-lsb order bin..bin+3
bcd  EQU temp0+4    ; temp4,5,6,7,8,9,10,11,12,13     10 BC digits or 10 ascii chars
pti  EQU temp0+14      ; pointers
pto  EQU temp0+15      ; pointers
ii   EQU temp0+16   
temp EQU temp0+17   
cnt  EQU temp0+18   

; Example Usage:
; call  b2bcd    ; convert to 32-bit binary to 10 bcd
; call  bcd2a    ; convert 10 bcd to 10 ascii
; goto  $


;******************************************************************
; Convert 32-bit binary number at <bin> into a bcd number
; at <bcd>. Uses Mike Keitz's procedure for handling bcd 
; adjust; Modified Microchip AN526 for 32-bits.

b2bcd   movlw d'32'    ; 32-bits
	movwf ii    ; make cycle counter
	clrf  bcd      ; clear result area
	clrf  bcd+1
	clrf  bcd+2
	clrf  bcd+3
	clrf  bcd+4
	
b2bcd2  movlw bcd      ; make pointer
	movwf FSR
	movlw 5
	movwf cnt

; Mike's routine:

b2bcd3  movlw 0x33     
	addwf INDF,f      ; add to both nybbles
	btfsc INDF,3      ; test if low result > 7
	andlw 0xf0     ; low result >7 so take the 3 out
	btfsc INDF,7      ; test if high result > 7
	andlw 0x0f     ; high result > 7 so ok
	subwf INDF,f      ; any results <= 7, subtract back
	incf  FSR,f    ; point to next
	decfsz   cnt
	goto  b2bcd3
	
	rlf   bin+3,f     ; get another bit
	rlf   bin+2,f
	rlf   bin+1,f
	rlf   bin+0,f
	rlf   bcd+4,f     ; put it into bcd
	rlf   bcd+3,f
	rlf   bcd+2,f
	rlf   bcd+1,f
	rlf   bcd+0,f
	decfsz   ii,f     ; all done?
	goto  b2bcd2      ; no, loop
	return         ; yes

;******************************************************************
; Convert the 10 binary coded digits (5 bytes) starting at 
; <bcd> into an ascii string also starting at <bcd>. Original
; bcd digits are lost.
; Ram Usage:
; CBLOCK   0x0c
; bin:4    ; 32-bit binary number (unsigned)
; bcd:10      ; 10 BC digits or 10 ascii chars
; pti,pto     ; pointers
; ii
; temp
; cnt
; ENDC

bcd2a   movlw bcd+9
	movwf pto      ; destination pointer
	movlw bcd+4
	movwf pti      ; source pointer
	movlw 5     ; 5 bytes to process
	movwf cnt

bcd2a1  movf  pti,w    ; get current input pointer
	movwf FSR
	decf  pti,f    ; prepare for next
	movf  INDF,w      ; get 2 bcds
	movwf temp     ; save for later
	movf  pto,w    ; get current output pointer
	movwf FSR
	decf  pto,f    ; prepare for next
	decf  pto,f
	movf  temp,w      ; get digits back
	andlw 0x0f     ; process lsd
	addlw "0"
	movwf INDF     ; to output
	decf  FSR,f
	swapf temp,w      ; process msd
	andlw 0x0f
	addlw "0"
	movwf INDF     ; to output
	decfsz   cnt      ; all digits?
	goto  bcd2a1
	return         ; yes

;==========================================================================
			END ; end of this program    
;==========================================================================

